#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This test shows a surrogate-assisted Bayesian calibration of a time dependent
    analytical function.

Author: Farid Mohammadi, M.Sc.
E-Mail: farid.mohammadi@iws.uni-stuttgart.de
Department of Hydromechanics and Modelling of Hydrosystems (LH2)
Institute for Modelling Hydraulic and Environmental Systems (IWS), University
of Stuttgart, www.iws.uni-stuttgart.de/lh2/
Pfaffenwaldring 61
70569 Stuttgart

Created on Fri Aug 9 2019

"""

import numpy as np
import pandas as pd
import sys
import joblib
# import bayesvalidrox
# Add BayesValidRox path
import sys
sys.path.append("../../src/bayesvalidrox/")

from pylink.pylink import PyLinkForwardModel
from surrogate_models.inputs import Input
from surrogate_models.surrogate_models import MetaModel
from post_processing.post_processing import PostProcessing
from bayes_inference.bayes_inference import BayesInference
from bayes_inference.discrepancy import Discrepancy
import matplotlib
matplotlib.use('agg')


if __name__ == "__main__":

    # Number of parameters
    ndim = 2  # 2, 10

    # =====================================================
    # =============   COMPUTATIONAL MODEL  ================
    # =====================================================
    Model = PyLinkForwardModel()

    Model.link_type = 'Function'
    Model.py_file = 'analytical_function'
    Model.name = 'AnalyticFunc'

    Model.Output.names = ['Z']

    # For Bayesian inversion synthetic data with X=[0,0]
    Model.observations = {}
    Model.observations['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.observations['Z'] = np.repeat([2.], 10)

    # For Checking with the MonteCarlo refrence
    Model.mc_reference = {}
    Model.mc_reference['Time [s]'] = np.arange(0, 10, 1.) / 9
    Model.mc_reference['mean'] = np.load(f"data/mean_{ndim}.npy")
    Model.mc_reference['std'] = np.load(f"data/std_{ndim}.npy")

    # =====================================================
    # =========   PROBABILISTIC INPUT MODEL  ==============
    # =====================================================
    # Define the uncertain parameters with their mean and
    # standard deviation
    Inputs = Input()

    # Assuming dependent input variables
    # Inputs.Rosenblatt = True

    for i in range(ndim):
        Inputs.add_marginals()
        Inputs.Marginals[i].name = "$\\theta_{"+str(i+1)+"}$"
        Inputs.Marginals[i].dist_type = 'uniform'
        Inputs.Marginals[i].parameters = [-5, 5]

    # arbitrary polynomial chaos
    # inputParams = np.load('data/InputParameters_{}.npy'.format(ndim))
    # for i in range(ndim):
    #     Inputs.add_marginals()
    #     Inputs.Marginals[i].name = f'$X_{i+1}$'
    #     Inputs.Marginals[i].input_data = inputParams[:, i]

    # =====================================================
    # ==========  DEFINITION OF THE METAMODEL  ============
    # =====================================================
    MetaModelOpts = MetaModel(Inputs)

    # Select if you want to preserve the spatial/temporal depencencies
    MetaModelOpts.dim_red_method = 'PCA'
    # MetaModelOpts.var_pca_threshold = 99.999
    # MetaModelOpts.n_pca_components = 10 #3, 10

    # Select your metamodel method
    # 1) PCE (Polynomial Chaos Expansion) 2) aPCE (arbitrary PCE)
    # 3) GPE (Gaussian Process Emulator)
    MetaModelOpts.meta_model_type = 'aPCE'

    # ------------------------------------------------
    # ------------- PCE Specification ----------------
    # ------------------------------------------------
    # Select the sparse least-square minimization method for
    # the PCE coefficients calculation:
    # 1)OLS: Ordinary Least Square  2)BRR: Bayesian Ridge Regression
    # 3)LARS: Least angle regression  4)ARD: Bayesian ARD Regression
    # 5)FastARD: Fast Bayesian ARD Regression
    # 6)VBL: Variational Bayesian Learning
    # 7)EBL: Emperical Bayesian Learning
    MetaModelOpts.pce_reg_method = 'FastARD'

    # Specify the max degree to be compared by the adaptive algorithm:
    # The degree with the lowest Leave-One-Out cross-validation (LOO)
    # error (or the highest score=1-LOO)estimator is chosen as the final
    # metamodel. pce_deg accepts degree as a scalar or a range.
    MetaModelOpts.pce_deg = np.arange(1, 12)

    # q-quasi-norm 0<q<1 (default=1)
    MetaModelOpts.pce_q_norm = 0.75 if ndim < 5 else 0.5
    # MetaModelOpts.pce_q_norm = np.linspace(0.25,0.75, 3)

    # Print summary of the regression results
    # MetaModelOpts.verbose = True

    # ------------------------------------------------
    # ------ Experimental Design Configuration -------
    # ------------------------------------------------
    MetaModelOpts.add_ExpDesign()

    # One-shot (normal) or Sequential Adaptive (sequential) Design
    MetaModelOpts.ExpDesign.method = 'sequential'
    MetaModelOpts.ExpDesign.n_init_samples = 5*ndim

    # Sampling methods
    # 1) random 2) latin_hypercube 3) sobol 4) halton 5) hammersley
    # 6) chebyshev(FT) 7) grid(FT) 8)user
    MetaModelOpts.ExpDesign.sampling_method = 'latin_hypercube'

    # Provide the experimental design object with a hdf5 file
    # MetaModelOpts.ExpDesign.hdf5_file = 'ExpDesign_AnalyticFunc.hdf5'

    # ------------------------------------------------
    # ------- Sequential Design configuration --------
    # ------------------------------------------------
    # Set the sampling parameters
    MetaModelOpts.ExpDesign.n_new_samples = 1
    MetaModelOpts.ExpDesign.n_max_samples = 75  # 150
    MetaModelOpts.ExpDesign.mod_LOO_threshold = 1e-16

    MetaModelOpts.adapt_verbose = True
    # 1) None 2) 'equal' 3)'epsilon-decreasing' 4) 'adaptive'
    MetaModelOpts.ExpDesign.tradeoff_scheme = 'adaptive'
    # MetaModelOpts.ExpDesign.n_replication = 5
    # -------- Exploration ------
    # 1)'Voronoi' 2)'random' 3)'latin_hypercube' 4)'LOOCV' 5)'dual annealing'
    MetaModelOpts.ExpDesign.explore_method = 'random'

    # Use when 'dual annealing' chosen
    MetaModelOpts.ExpDesign.max_func_itr = 100

    # Use when 'Voronoi' or 'random' or 'latin_hypercube' chosen
    MetaModelOpts.ExpDesign.n_canddidate = 10000
    MetaModelOpts.ExpDesign.n_cand_groups = 4

    # -------- Exploitation ------
    # 1)'BayesOptDesign' 2)'BayesActDesign' 3)'VarOptDesign' 4)'alphabetic'
    # 5)'Space-filling'
    MetaModelOpts.ExpDesign.exploit_method = 'VarOptDesign'

    # BayesOptDesign -> when data is available
    # 1) MI (Mutual information) 2) ALC (Active learning McKay)
    # 2)DKL (Kullback-Leibler Divergence) 3)DPP (D-Posterior-percision)
    # 4)APP (A-Posterior-percision)  # ['DKL', 'BME', 'infEntropy']
    # MetaModelOpts.ExpDesign.util_func = 'MI' #['MI', 'ALC']

    # VarBasedOptDesign -> when data is not available
    # 1)ALM 2)EIGF, 3)LOOCV
    # or a combination as a list
    MetaModelOpts.ExpDesign.util_func = 'ALM' #['ALM', 'EIGF']

    # alphabetic
    # 1)D-Opt (D-Optimality) 2)A-Opt (A-Optimality)
    # 3)K-Opt (K-Optimality) or a combination as a list
    # MetaModelOpts.ExpDesign.util_func = 'D-Opt'

    # Defining the measurement error, if it's known a priori
    obsData = pd.DataFrame(Model.observations, columns=Model.Output.names)
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2
    MetaModelOpts.Discrepancy = DiscrepancyOpts

    # Plot the posterior snapshots for SeqDesign
    MetaModelOpts.ExpDesign.post_snapshot = False
    MetaModelOpts.ExpDesign.step_snapshot = 1
    MetaModelOpts.ExpDesign.max_a_post = [0] * ndim

    # For calculation of validation error for SeqDesign
    prior = np.load(f"data/Prior_{ndim}.npy")
    prior_outputs = np.load(f"data/origModelOutput_{ndim}.npy")
    likelihood = np.load(f"data/validLikelihoods_{ndim}.npy")
    MetaModelOpts.valid_samples = prior
    MetaModelOpts.valid_model_runs = {'Z': prior_outputs}
    MetaModelOpts.valid_likelihoods = likelihood

    # >>>>>>>>>>>>>>>>>>>>>> Build Surrogate <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
    # Train the meta model
    PCEModel = MetaModelOpts.create_metamodel(Model)

    # Save PCE models
    with open(f'PCEModel_{Model.name}.pkl', 'wb') as output:
        joblib.dump(PCEModel, output, 2)

    # Load PCEModel
    # with open(f'PCEModel_{Model.name}.pkl', 'rb') as input:
    #     PCEModel = joblib.load(input)
    # import resource
    # m_1_out = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/1024
    # for i in range(40):
    #     PCEModel.input_obj.poly_coeffs_flag = False
    #     m_1 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/1024
    #     PCEModel = PCEModel.train_norm_design(Model, parallel=True)
    #     # y, y_std = PCEModel.eval_metamodel(samples=prior)
    #     y, y_std = PCEModel.eval_metamodel(nsamples=500*(i+1))
    #     m_2 = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/1024
    #     print(f"Memory itr {i+1}: I: {m_2-m_1:.2f} MB")
    # m_2_out = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss/1024
    # print(f"\nMemory all: I: {m_2_out-m_1_out:.2f} MB")
    # =====================================================
    # =========  POST PROCESSING OF METAMODELS  ===========
    # =====================================================
    PostPCE = PostProcessing(PCEModel)

    # Plot to check validation visually.
    PostPCE.valid_metamodel(n_samples=3)

    # Compute and print RMSE error
    PostPCE.check_accuracy(n_samples=300)

    # Compute the moments and compare with the Monte-Carlo reference
    if MetaModelOpts.meta_model_type != 'GPE':
        PostPCE.plot_moments()

    # Plot the evolution of the KLD,BME, and Modified LOOCV error
    if MetaModelOpts.ExpDesign.method == 'sequential':
        refBME_KLD = np.load("data/refBME_KLD_"+str(ndim)+".npy")
        PostPCE.plot_seq_design_diagnostics(refBME_KLD)

    # Plot the sobol indices
    if MetaModelOpts.meta_model_type != 'GPE':
        sobol_cell, total_sobol = PostPCE.sobol_indices()

    sys.exit()
    # =====================================================
    # ========  Bayesian inference with Emulator ==========
    # =====================================================
    BayesOpts = BayesInference(PCEModel)
    BayesOpts.emulator = True
    BayesOpts.plot_post_pred = True

    # BayesOpts.selected_indices = [0, 3, 5,  7, 9]
    # BME Bootstrap
    # BayesOpts.bootstrap = True
    # BayesOpts.n_bootstrap_itrs = 500
    # BayesOpts.bootstrap_noise = 100

    # Bayesian cross validation
    # BayesOpts.bayes_loocv = True

    # Select the inference method
    import emcee
    BayesOpts.inference_method = "MCMC"
    # Set the MCMC parameters passed to self.mcmc_params
    BayesOpts.mcmc_params = {
        'n_steps': 1e5,
        'n_walkers': 30,
        'moves': emcee.moves.KDEMove(),
        'multiprocessing': False,
        'verbose': False
        }

    # ----- Define the discrepancy model -------
    BayesOpts.measurement_error = obsData

    # # -- (Option B) --
    DiscrepancyOpts = Discrepancy('')
    DiscrepancyOpts.type = 'Gaussian'
    DiscrepancyOpts.parameters = obsData**2
    BayesOpts.Discrepancy = DiscrepancyOpts

    # -- (Option C) --
    # DiscOutputOpts = Input()
    # # # OutputName = 'Z'
    # DiscOutputOpts.add_marginals()
    # DiscOutputOpts.Marginals[0].Nnme = '$\sigma^2_{\epsilon}$'
    # DiscOutputOpts.Marginals[0].dist_type = 'uniform'
    # DiscOutputOpts.Marginals[0].parameters =  [0, 10]
    # BayesOpts.Discrepancy = {'known': DiscrepancyOpts,
    #                           'infer': Discrepancy(DiscOutputOpts)}

    # BayesOpts.bias_inputs = {'Z':np.arange(0, 10, 1.).reshape(-1,1) / 9}
    # DiscOutputOpts = Input()
    # # OutputName = 'lambda'
    # DiscOutputOpts.add_marginals()
    # DiscOutputOpts.Marginals[0].name = '$\lambda$'
    # DiscOutputOpts.Marginals[0].dist_type = 'uniform'
    # DiscOutputOpts.Marginals[0].parameters = [0, 1]

    # # OutputName = 'sigma_f'
    # DiscOutputOpts.add_marginals()
    # DiscOutputOpts.Marginals[1].Name = '$\sigma_f$'
    # DiscOutputOpts.Marginals[1].dist_type = 'uniform'
    # DiscOutputOpts.Marginals[1].parameters = [0, 1e-4]
    # BayesOpts.Discrepancy = Discrepancy(DiscOutputOpts)
    # BayesOpts.Discrepancy = {'known': DiscrepancyOpts,
    #                           'infer': Discrepancy(DiscOutputOpts)}
    # Start the calibration/inference
    Bayes_PCE = BayesOpts.create_inference()

    # Save class objects
    with open(f'Bayes_{Model.name}.pkl', 'wb') as output:
        joblib.dump(Bayes_PCE, output, 2)

    sys.exit('STOP!!')
    # =====================================================
    # ====== Bayesian inference with Forward Model ========
    # =====================================================
    if ndim < 7:
        PCEModel.ModelObj.Name = 'OrigAnalyticFunc'

        BayesOptsModel = bayesvalidrox.BayesInference(PCEModel)
        BayesOptsModel.emulator = False
        BayesOptsModel.NrofSamples = 10000

        # Evaluation of forward model predictions for the surrogate posterior
        # BayesOptsModel.Samples = Bayes_PCE.Samples
        # BayesOptsModel.SamplingMethod = 'MCMC'

        # Bootstrap for BME calulations
        BayesOptsModel.Bootstrap = True
        BayesOptsModel.BootstrapItrNr = 1
        BayesOptsModel.BootstrapNoise = 0.05

        BayesOptsModel.PlotPostDist = True
        BayesOptsModel.PlotPostPred = False

        # ----- Define the discrepancy model -------
        BayesOptsModel.MeasurementError = obsData
        # (Option B)
        DiscrepancyOpts = bayesvalidrox.Discrepancy('')
        DiscrepancyOpts.Type = 'Gaussian'
        DiscrepancyOpts.Parameters = obsData ** 2
        BayesOptsModel.Discrepancy = DiscrepancyOpts

        # Start the infernce
        Bayes_Model = BayesOptsModel.create_Inference()

    # =====================================================
    # ==============  Save class objects  =================
    # =====================================================
    if ndim < 7:
        with open('AnalyticFunc_Results.pkl', 'wb') as output:
            joblib.dump(Bayes_Model, output, 2)
